package net.npaka.wifidirectex;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;

//チャットマネージャ
public class ChatManager {
    //状態定数
    public static final int STATE_NONE       = 0;
    public static final int STATE_LISTEN     = 1;
    public static final int STATE_CONNECTING = 2;
    public static final int STATE_CONNECTED  = 3;

    //変数
    private WiFiDirectEx    parent;    //本体
    private int             state;     //状態
    private AcceptThread    acceptT;   //サーバの接続待ちスレッド
    private ConnectThread   connectT;  //クライアントの接続要求スレッド
    private ConnectedThread connectedT;//接続中の処理スレッド
    
    //コンストラクタ
    public ChatManager(WiFiDirectEx parent) {
        this.parent = parent;
        this.state  = STATE_NONE;
    }

    //状態の指定
    private synchronized void setState(final int state) {
        this.state = state;
        if (state == STATE_CONNECTED) {
            parent.setState("接続完了");
        } else if (state == STATE_CONNECTING) {
            parent.setState("接続中");
        } else if (state == STATE_LISTEN) {
            parent.setState("接続待ち");
        } else if (state == STATE_NONE) {
            parent.setState("未接続");
        }
    }

    //状態の取得
    public synchronized int getState() {
        return state;
    }

    //サーバの接続待ちスレッドの開始
    public synchronized void accept() {
        if (connectT   != null) {connectT.cancel();  connectT   = null;}
        if (connectedT != null) {connectedT.cancel();connectedT = null;}
        if (acceptT == null) {
            acceptT = new AcceptThread();
            acceptT.start();
            setState(STATE_LISTEN);
        }
    }    
    
    //クライアントの接続要求スレッドの開始
    public synchronized void connect(String host) {
        if (acceptT    != null) {acceptT.cancel();   acceptT    = null;}
        if (connectedT != null) {connectedT.cancel();connectedT = null;}
        if (connectT == null) {
            connectT = new ConnectThread(host);
            connectT.start();
            setState(STATE_CONNECTING);
        }
    }    
    
    //接続中の処理スレッドの開始
    public synchronized void connected(Socket socket) {
        if (acceptT  != null) {acceptT.cancel(); acceptT  = null;}
        if (connectT != null) {connectT.cancel();connectT = null;}
        if (connectedT == null) {
            connectedT = new ConnectedThread(socket);
            connectedT.start();
            setState(STATE_CONNECTED);
        }
    }

    //送信データの書き込み
    public synchronized void write(byte[] out) {
        if (state != STATE_CONNECTED) return;
        connectedT.write(out);
    }
    
    //スレッドの停止
    public synchronized void stop() {
        if (connectT   != null) {connectT.cancel();  connectT   = null;}
        if (connectedT != null) {connectedT.cancel();connectedT = null;}
        if (acceptT    != null) {acceptT.cancel();   acceptT    = null;}
        setState(STATE_NONE);
    }

    //サーバの接続待ちスレッド(10)
    private class AcceptThread extends Thread {
        private ServerSocket serverSocket;
        
        //処理
        public void run() {
            parent.addText("サーバ接続待ちの開始");
            Socket socket = null;
            try {
                serverSocket = new ServerSocket(8988);
            } catch (Exception e) {
                parent.addText("サーバ接続待ちのエラー>"+e.toString());
                setState(STATE_NONE);
            }            
            while (state != STATE_CONNECTED) {
                try {
                    socket = serverSocket.accept();
                } catch (Exception e) {
                    break;
                }
                if (socket != null) {
                    if (state == STATE_LISTEN || state == STATE_CONNECTING) {
                        connected(socket);
                    } else if (state == STATE_NONE || state == STATE_CONNECTED) {
                        try {
                            socket.close();
                        } catch (Exception e) {
                        }
                    }
                }
            }
            parent.addText("サーバ接続待ちの完了");
        }
        
        //キャンセル
        public void cancel() {
            try {
                serverSocket.close();
            } catch (Exception e) {
            }
        }
    }
    
    //クライアントの接続要求スレッド(11)
    private class ConnectThread extends Thread {
        private String host;
        private Socket socket;
        private boolean cancel;

        //コンストラクタ
        public ConnectThread(String host) {
            this.host = host;
        }

        //処理
        public void run() {
            parent.addText("クライアント接続要求の開始>"+host);
            try {
                socket = new Socket(host, 8988);
                connectT = null;
                connected(socket);
            } catch (Exception e) {
                parent.addText("クライアント接続要求の完了");
                
                //停止
                try {
                    socket.close();
                } catch (Exception e2) {
                }
                setState(STATE_NONE);
                
                //サーバの再開
                if (!cancel) accept();
            }
        }

        //キャンセル
        public void cancel() {
            cancel = true;
            try {
                socket.close();
            } catch (Exception e) {
            }
        }
    }    
    
    //接続中の処理スレッド(12)
    private class ConnectedThread extends Thread {
        private Socket       socket;
        private InputStream  in;
        private OutputStream out;
        private boolean      cancel;

        //コンストラクタ
        public ConnectedThread(Socket socket) {
            parent.addText("接続中処理の開始");
            try {
                this.socket = socket;
                this.in     = socket.getInputStream();
                this.out    = socket.getOutputStream();
            } catch (Exception e) {
                parent.addText("接続中処理のエラー>"+e.getMessage());
                setState(STATE_NONE);
            }
        }

        //処理
        public void run() {
            byte[] buf = new byte[1024];
            while (true) {
                try {
                    int size = in.read(buf);
                    final String str = new String(buf, 0, size);
                    parent.addText(str);
                } catch (Exception e) {
                    parent.addText("接続中処理の完了");
                    //停止
                    try {
                        socket.close();
                    } catch (Exception e2) {
                    }
                    setState(STATE_NONE);
                    
                    //サーバの再開
                    if (!cancel) accept();
                    break;
                }
            }
        }

        //書き込み
        public void write(byte[] buf) {
            try {
                out.write(buf);
            } catch (Exception e) {
            }
        }

        //キャンセル
        public void cancel() {
            cancel = true;
            try {
                socket.close();
            } catch (Exception e) {
            }
        }
    }
}
